/*********************************************************************/
/*      Licensed Materials - Property of IBM                         */
/*      "Restricted Materials of IBM"                                */
/*      5748-T15                                                     */
/*      (C) Copyright IBM Corp. 2006                                 */
/*      All Rights Reserved                                          */
/*      US Government Users Restricted Rights - Use, duplication     */
/*      or dislosure restricted by GSA ADP Schedule Contract         */
/*      with IBM Corp.                                               */
/*********************************************************************/
/* This file is provided on an "as is" basis.  There are no          */
/* warranties, express or implied, including the implied warranties  */
/* of merchantability and fitness for a particular purpose.          */
/* IBM has no obligation to provide service, defect correction, or   */
/* any maintenance for the file.  IBM has no obligation to supply    */
/* any updates or enhancements for the file to you even if such are  */
/* or later become available.                                        */
/*********************************************************************/
/* Change History                                                    */
/*                                                                   */
/*                                                                   */
/*                                                                   */
/*********************************************************************/
/* GENERAL INFORMATION:                                              */
/*                                                                   */
/* Header Name:   qm5a70.h                                           */
/*                                                                   */
/* Description:   Header file used by MEM5 driver segments.          */
/*                It contains prototypes and various #defines.       */
/*                                                                   */
/*********************************************************************/
#ifndef __qm5a70_h__
#define __qm5a70_h__

#define prterr(x) wtopc(x, WTOPC_EBROUT, WTOPC_NO_CHAIN, WTOPC_NO_HEADER)
#define prtenderr(x) wtopc(x, WTOPC_EBROUT, WTOPC_NO_CHAIN, WTOPC_NO_HEADER)
#define prtmsg(x) if ( ecbptr()->ebcm01 == TRUE ) wtopc(x, WTOPC_EBROUT, WTOPC_NO_CHAIN, WTOPC_NO_HEADER); else
#define prtendmsg(x) if ( ecbptr()->ebcm01 == TRUE ) wtopc(x, WTOPC_EBROUT, WTOPC_NO_CHAIN_END, WTOPC_NO_HEADER); else

struct permutations {
    char *desc;           // this is what will print for "verbose"
    char *owner;          // owner string to pass to gsysc/rsysc
    int   flags;          // flags for the MEM5 driver and gsysc/rsysc
    unsigned char  debug; // print each gsysc/rsysc if DEBUG
};

#define verboseFlag (ecbptr()->ebcm01)
#define errorFlag   (ecbptr()->ebx000)
#define Verbose     (verboseFlag == TRUE)
#define DoErrors    (errorFlag   == TRUE)

/* system heap flag settings used by the MEM5 driver */
#define TPF41_API      0x00
#define TPF51_API      0x01
#define DO_64BIT       0x02
#define MB_FRAMES      0x04
#define UNIQUE_TOKEN   0x08
#define ERROR_PATH     0x10

#define USE_1MEG_FRAMES  (flags & MB_FRAMES)
#define USE_64BIT        (flags & DO_64BIT)
#define USE_UNIQUE_TOKEN (flags & UNIQUE_TOKEN)
#define USE_TPF51_API    (flags & (TPF51_API|MB_FRAMES|DO_64BIT|UNIQUE_TOKEN))
#define EXPECTED_ERROR   (flags & ERROR_PATH)

#define DEBUG    TRUE
#define NO_DEBUG FALSE

#define SMALL_NBR_OF_4K_FRAMES    3
#define MEDIUM_NBR_OF_4K_FRAMES   9
#define LARGE_NBR_OF_4K_FRAMES    20

#define SMALL_NBR_OF_1MEG_FRAMES  2
#define MEDIUM_NBR_OF_1MEG_FRAMES 5
#define LARGE_NBR_OF_1MEG_FRAMES  14

#define FOUR_KB 4096
#define ONE_MEG 1048576

#define ETPFGSYSCTOKEN 0
#define ETPFRSYSCADDR  1
#define ETPFRSYSCFRAME 2
#define ETPFRSYSCTOKEN 3

#define LEVEL_2    0x80
#define LEVEL_3    0x40
#define LEVEL_7    0x20
#define LEVEL_9    0x10
#define LEVEL_C    0x08
#define LEVEL_F    0x04

#define SIZE_1K   1024
#define SIZE_4K   (4 * SIZE_1K)
#define SIZE_64K  (64 * SIZE_1K)
#define SIZE_256K (256 * SIZE_1K)

typedef enum _task {USE_SYSTEM_HEAP, FREE_ECB_HEAP, REALLOC_ECB_HEAP, MALLOC_ECB_HEAP, MAX_ECB} t_task;
typedef  enum t_heap {TPF_HEAP_31=0, TPF_HEAP_64} t_heap;

struct PassedParms {
    t_task task;
    unsigned char  verboseFlagSetting;
    unsigned char  errorFlagSetting;
    void *genericPtr;
    int genericNum;
    int genericSize;
};

struct SystemHeapParms {
    int  nbrOfFrames;
    int  flags;
    char token[9];
};

typedef struct ECBHeapParms {
    int size;
    int numElements;
}ECBHEAPPARMS;

void  CreateECB(t_task);
void  CreateECB_void(t_task, void *);
void  CreateECB_int(t_task, int);
void  CreateECB_2int(t_task, int, int);
void  prterrno(int);
void  PrintMsg(char *);
void  PrintError(char *);
void *GetSystemHeap(int, char *, char *, int);
void *GetSystemHeapTPF41API(int, char *, int);
void *GetSystemHeapTPF51API(int, char *, char *, int);
int   ReturnSystemHeap(void *, int, char *, char *, int);
int   WriteToSystemHeap(void *, int, int);

#endif /* ndef __qm5a70_h__ */
