/*********************************************************************/
/*      Licensed Materials - Property of IBM                         */
/*      "Restricted Materials of IBM"                                */
/*      5748-T15                                                     */
/*      (C) Copyright IBM Corp. 2006                                 */
/*      All Rights Reserved                                          */
/*      US Government Users Restricted Rights - Use, duplication     */
/*      or dislosure restricted by GSA ADP Schedule Contract         */
/*      with IBM Corp.                                               */
/*********************************************************************/
/* This file is provided on an "as is" basis.  There are no          */
/* warranties, express or implied, including the implied warranties  */
/* of merchantability and fitness for a particular purpose.          */
/* IBM has no obligation to provide service, defect correction, or   */
/* any maintenance for the file.  IBM has no obligation to supply    */
/* any updates or enhancements for the file to you even if such are  */
/* or later become available.                                        */
/*********************************************************************/
/* Change History                                                    */
/*                                                                   */
/*                                                                   */
/*                                                                   */
/*********************************************************************/
/* GENERAL INFORMATION:                                              */
/*                                                                   */
/* Segment Name:  qm5a80.c                                           */
/*                                                                   */
/* Abstract:      Member of the QM5A CSO which comprises the         */
/*                TPF GCC MEM5 driver.                               */
/*                                                                   */
/* Description:   This segment contains System Heap driver tests     */
/*                                                                   */
/*********************************************************************/
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <tpf/tpfapi.h>
#include <tpf/c_eb0eb.h>
#include <tpf/sysapi.h>
#include <unistd.h> // for sleep()
#include <limits.h> // for INT_MAX #define
#include <errno.h>
#include "qm5a70.h"

extern const char *memPgm;
extern char  *msgPrefix;
extern char  msgBuffer[ ];
extern struct permutations perm[ ];
extern int nbr_of_perms;

char  systemHeapAbortMsg[] =
         "skipping further system heap tests due to errors";

//---------------------------------------------------------------------
// function prototypes
//---------------------------------------------------------------------
int   DoSystemHeapNormalPaths(char *, int);
int   DoSystemHeapErrorPaths(char *, int);
int   FindSystemHeap(char *, void *, int);
void  BuildTokenString(char *, char *);

/*********************************************************************/
/* Function Name:                                                    */
/*                                                                   */
/* Number of Cases:                                                  */
/*                                                                   */
/*********************************************************************/

void systemheap_func(void)
{
    int i, rc;
    unsigned char savedVerboseFlag;

    msgPrefix = "qm5a";
    PrintMsg("entering systemheap_func");
    savedVerboseFlag = verboseFlag;

    for (i = 0; i < nbr_of_perms; i++) {
        if (DoErrors) {
            sprintf(msgBuffer,
                    "Doing system heap error tests for %s",
                    perm[i].desc);
        } else {
            sprintf(msgBuffer,
                    "Doing system heap test cases for %s",
                    perm[i].desc);
        }
        PrintMsg(msgBuffer);
        PrintMsg("End of Display");
        verboseFlag &= perm[i].debug;
        if (DoErrors) {
            rc = DoSystemHeapErrorPaths(perm[i].owner, perm[i].flags);
        } else {
            rc = DoSystemHeapNormalPaths(perm[i].owner, perm[i].flags);
        }
        verboseFlag = savedVerboseFlag;
        if (rc != 0) {
            verboseFlag = savedVerboseFlag;
            PrintError(systemHeapAbortMsg);
            return;
        }

    }

    PrintMsg("completed systemheap_func");

    return;
}

void systemheap2_func(void)
{
    void *systemHeapPtr;
    struct SystemHeapParms *SHP;
    int i;
    unsigned char savedVerboseFlag;
#   define TOKENID "abcdefgh"
#   define NBROFFRAMES 2

    msgPrefix = "qm5a";
    PrintMsg("entering systemheap2_func");
    savedVerboseFlag = verboseFlag;

    // Have one ECB request system heap and a second ECB
    // access, modify, and release it

    for (i = 0; i < nbr_of_perms; i++) {
        if (DoErrors) {
            sprintf(msgBuffer,
                    "Doing double ECB system heap error tests for %s",
                    perm[i].desc);

        } else {
            sprintf(msgBuffer,
                    "Doing double ECB system heap test cases"
                        " for %s",
                    perm[i].desc);
        }
        PrintMsg(msgBuffer);

        verboseFlag &= perm[i].debug;
        // get some system heap to start with
        systemHeapPtr = GetSystemHeap(NBROFFRAMES,
                                      TOKENID,
                                      perm[i].owner,
                                      perm[i].flags);
        if (systemHeapPtr) {
            SHP = (struct SystemHeapParms *)systemHeapPtr;
            cinfc(CINFC_WRITE,CINFC_CMMCTK2);
            SHP->nbrOfFrames = NBROFFRAMES;
            SHP->flags = perm[i].flags;
            memcpy(SHP->token, TOKENID, 8);
            keyrc();
            // create the 2nd ECB to do the actual tests
            CreateECB_void(USE_SYSTEM_HEAP, SHP);
            verboseFlag = savedVerboseFlag;
            if (DoErrors) {
                // This error path dumps so try to reduce the impact:
                sleep(2);  // slow it down a bit
                i += 2;    // and only do every third case
            }
        } else {
            verboseFlag = savedVerboseFlag;
            PrintError(systemHeapAbortMsg);
            return;
        }
    }

    PrintMsg("completed systemheap2_func");

    return;
}

int DoSystemHeapNormalPaths(char *owner, int flags)
{
    void *systemHeapPtrA, *systemHeapPtrB;
    char *tokenA, *tokenB;
    char tokenChar[10] = "123456789";
    int  nbrOfFramesA, nbrOfFramesB;
    int  rc, previousError = FALSE;

    // test case: gsysc with 9 byte token (only 8 of which are used)
    nbrOfFramesA = 2;
    systemHeapPtrA = GetSystemHeap(nbrOfFramesA, tokenChar, owner, flags);
    if (!systemHeapPtrA) {
        PrintError("gsysc failed with 9 byte token");
        return(1);
    }
    tokenChar[8] = 'X'; // change token byte 9 to make sure it isn't being used
    if (FindSystemHeap(tokenChar, systemHeapPtrA, flags) != 0) {
        PrintError("tpf_fsysc failed with 9 byte token");
        previousError = TRUE;
    }
    rc = ReturnSystemHeap(systemHeapPtrA,
                          nbrOfFramesA,
                          tokenChar,
                          owner,
                          flags);
    if ((rc != RSYSC_OK) || (previousError)) {
        return(1);
    }

    // test case: gsysc with token with embedded null
    nbrOfFramesA = 1;
    tokenChar[8] = '\0'; // make token 8 bytes int
    tokenChar[2] = '\0'; // and embed a null
    systemHeapPtrA = GetSystemHeap(nbrOfFramesA, tokenChar, owner, flags);
    if (!systemHeapPtrA) {
        PrintError("gsysc failed with token containing embedded null");
        return(1);
    }
    if (FindSystemHeap(tokenChar, systemHeapPtrA, flags) != 0) {
        PrintError("tpf_fsysc failed with token containing embedded null");
        previousError = TRUE;
    }
    rc = ReturnSystemHeap(systemHeapPtrA,
                          nbrOfFramesA,
                          tokenChar,
                          owner,
                          flags);
    if ((rc != RSYSC_OK) || (previousError)) {
        return(1);
    }

    // test case: gsysc with token with initial null
    nbrOfFramesA = 1;
    tokenChar[2] = '3';  // get rid of previous embedded null
    tokenChar[0] = '\0'; // and make 1st byte null
    systemHeapPtrA = GetSystemHeap(nbrOfFramesA, tokenChar, owner, flags);
    if (!systemHeapPtrA) {
        PrintError("gsysc failed with token containing initial null");
        return(1);
    }
    if (FindSystemHeap(tokenChar, systemHeapPtrA, flags) != 0) {
        PrintError("tpf_fsysc failed with token containing initial null");
        previousError = TRUE;
    }
    rc = ReturnSystemHeap(systemHeapPtrA,
                          nbrOfFramesA,
                          tokenChar,
                          owner,
                          flags);
    if ((rc != RSYSC_OK) || (previousError)) {
        return(1);
    }

    // test case: gsysc with token of 8 nulls
    memset(tokenChar, 0x00, sizeof(tokenChar));
    nbrOfFramesA = 3;
    systemHeapPtrA = GetSystemHeap(nbrOfFramesA, tokenChar, owner, flags);
    if (!systemHeapPtrA) {
        PrintError("gsysc failed when given token of 8 nulls");
        return(1);
    }
    if (FindSystemHeap(tokenChar, systemHeapPtrA, flags) != 0) {
        PrintError("tpf_fsysc failed when given token of 8 nulls");
        previousError = TRUE;
    }
    rc = ReturnSystemHeap(systemHeapPtrA,
                          nbrOfFramesA,
                          tokenChar,
                          owner,
                          flags);
    if ((rc != RSYSC_OK) || (previousError)) {
        return(1);
    }

    // test case: get 1 frame with a non-unique token
    nbrOfFramesA = 1;
    tokenA = "TokenID1";
    systemHeapPtrA = GetSystemHeap(nbrOfFramesA, tokenA, owner,
                                   (flags & ~UNIQUE_TOKEN));
    if (!systemHeapPtrA) {
        PrintError("gsysc failed on 1 frame request with non-unique token");
        return(1);
    }

    // test case: get, find, & release another frame with same token
    systemHeapPtrB = GetSystemHeap(nbrOfFramesA, tokenA, owner, flags);
    if (!systemHeapPtrB) {
        PrintError("gsysc failed when asked for frames with duplicate token");
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         (flags & ~UNIQUE_TOKEN));
        return(1);
    }
    if (FindSystemHeap(tokenA, systemHeapPtrB, flags) != 0) {
        PrintError("tpf_fsysc failed when sent duplicate token");
        previousError = TRUE;
    }
    rc = ReturnSystemHeap(systemHeapPtrB,
                          nbrOfFramesA,
                          tokenA,
                          owner,
                          flags);
    if ((rc != RSYSC_OK) || (previousError)) {
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         (flags & ~UNIQUE_TOKEN));
        return(1);
    }

    // write to original system heap request
    rc = WriteToSystemHeap(systemHeapPtrA, nbrOfFramesA, flags);
    if (rc) {
        // write failed
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         (flags & ~UNIQUE_TOKEN));
        return(1);
    }
    defrc();

    // test case: get, write to, and release a small number of frames
    if (USE_1MEG_FRAMES) {
        nbrOfFramesB = SMALL_NBR_OF_1MEG_FRAMES;
    } else {
        nbrOfFramesB = SMALL_NBR_OF_4K_FRAMES;
    }
    tokenB = "TokenSno";
    systemHeapPtrB = GetSystemHeap(nbrOfFramesB, tokenB, owner, flags);
    if (!systemHeapPtrB) {
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         (flags & ~UNIQUE_TOKEN));
        return(1);
    }
    rc = WriteToSystemHeap(systemHeapPtrB, nbrOfFramesB, flags);
    if (rc) {
        // write failed
        previousError = TRUE;
    }
    defrc();

    // find token
    if (!previousError) {
        if (FindSystemHeap(tokenB, systemHeapPtrB, flags) != 0) {
            PrintError("tpf_fsysc failed on small system heap amount");
            previousError = TRUE;
        }
    }
    // return the system heap using TPF4.1 rsysc
    rc = ReturnSystemHeap(systemHeapPtrB,
                          nbrOfFramesB,
                          tokenB,
                          "different owner string",
                          TPF41_API);
    if ((rc != RSYSC_OK) || (previousError)) {
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         (flags & ~UNIQUE_TOKEN));
        return(1);
    }

    // test case: get, write to, and release a large number of frames
    if (USE_1MEG_FRAMES) {
        nbrOfFramesB = LARGE_NBR_OF_1MEG_FRAMES;
    } else {
        nbrOfFramesB = LARGE_NBR_OF_4K_FRAMES;
    }
    tokenB = "Token2  ";
    systemHeapPtrB = GetSystemHeap(nbrOfFramesB, tokenB, owner, flags);
    if (!systemHeapPtrB) {
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         (flags & ~UNIQUE_TOKEN));
        return(1);
    }
    rc = WriteToSystemHeap(systemHeapPtrB, nbrOfFramesB, flags);
    if (rc) {
        // write failed
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         (flags & ~UNIQUE_TOKEN));
        ReturnSystemHeap(systemHeapPtrB,
                         nbrOfFramesB,
                         tokenB,
                         owner,
                         flags);
        return(1);
    }
    defrc();

    rc = ReturnSystemHeap(systemHeapPtrB,
                          nbrOfFramesB,
                          tokenB,
                          owner,
                          flags);
    if (rc != RSYSC_OK) {
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         (flags & ~UNIQUE_TOKEN));
        return(1);
    }

    // test case: get, and write to, a medium number of frames
    if (USE_1MEG_FRAMES) {
        nbrOfFramesB = MEDIUM_NBR_OF_1MEG_FRAMES;
    } else {
        nbrOfFramesB = MEDIUM_NBR_OF_4K_FRAMES;
    }
    tokenB = "Token2  ";
    systemHeapPtrB = GetSystemHeap(nbrOfFramesB, tokenB, owner, flags);
    if (!systemHeapPtrB) {
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         (flags & ~UNIQUE_TOKEN));
        return(1);
    }
    // write to this system heap
    rc = WriteToSystemHeap(systemHeapPtrB, nbrOfFramesB, flags);
    if (rc) {
        // write failed
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         (flags & ~UNIQUE_TOKEN));

        ReturnSystemHeap(systemHeapPtrB,
                         nbrOfFramesB,
                         tokenB,
                         owner,
                         flags);
        return(1);
    }
    defrc();

    // return original frame request
    rc = ReturnSystemHeap(systemHeapPtrA,
                          nbrOfFramesA,
                          tokenA,
                          owner,
                          (flags & ~UNIQUE_TOKEN));
    if (rc != RSYSC_OK) {
        ReturnSystemHeap(systemHeapPtrB,
                         nbrOfFramesB,
                         tokenB,
                         owner,
                         flags);
        return(1);
    }

    // return last outstanding frame requested
    rc = ReturnSystemHeap(systemHeapPtrB,
                         nbrOfFramesB,
                         tokenB,
                         owner,
                         flags);
    if (rc != RSYSC_OK) {
        return(1);
    }

    return(0);

}


int DoSystemHeapErrorPaths(char *owner, int flags)
{
    void *systemHeapPtrA, *systemHeapPtrB, *systemHeapPtrC;
    char *tokenA, *tokenB, *tokenC;
    int  nbrOfFramesA, nbrOfFramesB, nbrOfFramesC, rc;
    char tokenChar[9] = "abcdefgh";

    // error test case: gsysc with zero frames requested
    PrintMsg("attempted gsysc with zero frames");
    tokenA = "foobar17";
    nbrOfFramesA = 0;
    systemHeapPtrA = GetSystemHeap(nbrOfFramesA, tokenA, owner,
                                   flags|ERROR_PATH);
    if (systemHeapPtrA) {
        PrintError("gsysc failed to fail with zero frames");
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         flags);
        return(1);
    }

    // error test case: gsysc with -1 frames requested
    PrintMsg("attempted gsysc with -1 frames");
    tokenA = "foobar17";
    nbrOfFramesA = -1;
    systemHeapPtrA = GetSystemHeap(nbrOfFramesA, tokenA, owner,
                                   flags|ERROR_PATH);
    if (systemHeapPtrA) {
        PrintError("gsysc failed to fail with -1 frames");
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         flags);
        return(1);
    }

    // error test case: gsysc with max int for frames value
    PrintMsg("attempted gsysc with max int for frames value");
    tokenA = "maximum ";
    nbrOfFramesA = INT_MAX;
    systemHeapPtrA = GetSystemHeap(nbrOfFramesA, tokenA, owner,
                                   flags|ERROR_PATH);

    if (systemHeapPtrA) {
        PrintError("gsysc failed to fail with max int for frames value");
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         flags);
        return(1);
    }

    // error test case: tpf_fsysc with nonexistent token
    PrintMsg("attempted tpf_fsysc with nonexistent token");
    tokenA = "avcWx9iP";
    if  (FindSystemHeap(tokenA, (void *)0, flags|ERROR_PATH) == 0){
        sprintf(msgBuffer,
            "tpf_fsysc failed to fail with nonexistent token \"%s\"",
            tokenA);
        PrintError(msgBuffer);
        return(1);

    }

    // error test case: tpf_fsysc with invalid token pointer
    PrintMsg("attempted tpf_fsysc with token pointer of -1");
    if (FindSystemHeap((char *)-1, (void *)0, flags|ERROR_PATH) == 0) {
        PrintError("tpf_fsysc failed to fail with token pointer of -1");
        return(1);
    }

//  Get some heap for real to execute more complicated error paths.
//  Note: A & B have the same number of frames but different tokens
//        B & C have the same token but a different number of frames
//              (which is why C is always forced down the non-unique path)
    nbrOfFramesA = 2;
    tokenA = "tokenA  ";
    systemHeapPtrA = GetSystemHeap(nbrOfFramesA, tokenA, owner, flags);
    if (!systemHeapPtrA) {
        return(1);
    }

    nbrOfFramesB = nbrOfFramesA;
    tokenB = "tokenIsB";
    systemHeapPtrB = GetSystemHeap(nbrOfFramesB, tokenB, owner, flags);
    if (!systemHeapPtrB) {
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         flags);
        return(1);
    }

    if (USE_UNIQUE_TOKEN) {
    // error test case: gsysc twice with the same unique token
        PrintMsg("attempted gsysc twice with same unique token");
        systemHeapPtrC = GetSystemHeap(nbrOfFramesB, tokenB, owner, flags|ERROR_PATH);
        if (systemHeapPtrC) {
            PrintError("gsysc failed to fail when given non-unique token");
            ReturnSystemHeap(systemHeapPtrA,
                             nbrOfFramesA,
                             tokenA,
                             owner,
                             flags);
            ReturnSystemHeap(systemHeapPtrB,
                             nbrOfFramesB,
                             tokenB,
                             owner,
                             flags & ~UNIQUE_TOKEN);
            ReturnSystemHeap(systemHeapPtrC,
                             nbrOfFramesB,
                             tokenB,
                             owner,
                             flags & ~UNIQUE_TOKEN);
            return(1);
        }
    }

    // get systemHeapPtrC for real this time
    nbrOfFramesC = nbrOfFramesA + 1;
    tokenC = tokenB;
    systemHeapPtrC = GetSystemHeap(nbrOfFramesC, tokenC, owner,
                                   flags & ~UNIQUE_TOKEN);
    if (!systemHeapPtrC) {
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         flags);
        ReturnSystemHeap(systemHeapPtrB,
                         nbrOfFramesB,
                         tokenB,
                         owner,
                         flags);
        return(1);
    }


    // error test case: tpf_rsysc with an invalid address - 1
    PrintMsg("attempted tpf_rsysc with address of zero");
    rc = ReturnSystemHeap((void *)1 , nbrOfFramesA, tokenA, owner,
                          (flags | ERROR_PATH) & ~UNIQUE_TOKEN);
    if (rc == RSYSC_OK) {
        PrintError("tpf_rsysc failed to fail when given address of zero");
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         flags);
        ReturnSystemHeap(systemHeapPtrB,
                         nbrOfFramesB,
                         tokenB,
                         owner,
                         flags);
        ReturnSystemHeap(systemHeapPtrC,
                         nbrOfFramesC,
                         tokenC,
                         owner,
                         flags & ~UNIQUE_TOKEN);
        return(1);
    }

    // error test case: tpf_rsysc with address of -1
    PrintMsg("attempted tpf_rsysc with address of -1");
    rc = ReturnSystemHeap((void *)-1, nbrOfFramesA, tokenA, owner,
                          (flags | ERROR_PATH) & ~UNIQUE_TOKEN);
    if (rc == RSYSC_OK) {
        PrintError("tpf_rsysc failed to fail when given address of -1");
        ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         flags);
        ReturnSystemHeap(systemHeapPtrB,
                         nbrOfFramesB,
                         tokenB,
                         owner,
                         flags);
        ReturnSystemHeap(systemHeapPtrC,
                         nbrOfFramesC,
                         tokenC,
                         owner,
                         flags & ~UNIQUE_TOKEN);
        return(1);
    }

    // error test case: tpf_rsysc with the wrong address
    PrintMsg("attempted tpf_rsysc with the wrong address");
    rc = ReturnSystemHeap(systemHeapPtrA, nbrOfFramesB, tokenB, owner,
                          (flags | ERROR_PATH) & ~UNIQUE_TOKEN);
    if (rc == RSYSC_OK) {
        PrintError("tpf_rsysc failed to fail when given the wrong address");
        ReturnSystemHeap(systemHeapPtrB,
                         nbrOfFramesB,
                         tokenB,
                         owner,
                         flags);
        ReturnSystemHeap(systemHeapPtrC,
                         nbrOfFramesC,
                         tokenC,
                         owner,
                         flags & ~UNIQUE_TOKEN);
        return(1);
    }

    // error test case: tpf_rsysc with null token (when original had one)
    PrintMsg("attempted tpf_rsysc with null token");
    memset(tokenChar, 0x00, sizeof(tokenChar));
    rc = ReturnSystemHeap(systemHeapPtrA, nbrOfFramesA, tokenChar, owner,
                          (flags | ERROR_PATH) & ~UNIQUE_TOKEN);
    if (rc == RSYSC_OK) {
        PrintError("tpf_rsysc failed to fail when given null token");
        ReturnSystemHeap(systemHeapPtrB,
                         nbrOfFramesB,
                         tokenB,
                         owner,
                         flags);
        ReturnSystemHeap(systemHeapPtrC,
                         nbrOfFramesC,
                         tokenC,
                         owner,
                         flags & ~UNIQUE_TOKEN);
        return(1);
    }

    // error test case: tpf_rsysc with the wrong token
    PrintMsg("attempted tpf_rsysc with the wrong token");
    rc = ReturnSystemHeap(systemHeapPtrA, nbrOfFramesA, tokenB, owner,
                          (flags | ERROR_PATH) & ~UNIQUE_TOKEN);
    if (rc == RSYSC_OK) {
        PrintError("tpf_rsysc failed to fail when given the wrong token");
        ReturnSystemHeap(systemHeapPtrB,
                         nbrOfFramesB,
                         tokenB,
                         owner,
                         flags);
        ReturnSystemHeap(systemHeapPtrC,
                         nbrOfFramesC,
                         tokenC,
                         owner,
                         flags & ~UNIQUE_TOKEN);
        return(1);
    }

    // error test case: tpf_rsysc the same address twice
    rc = ReturnSystemHeap(systemHeapPtrA,
                         nbrOfFramesA,
                         tokenA,
                         owner,
                         flags);
    if (rc != RSYSC_OK) {
        ReturnSystemHeap(systemHeapPtrB,
                         nbrOfFramesB,
                         tokenB,
                         owner,
                         flags);
        ReturnSystemHeap(systemHeapPtrC,
                         nbrOfFramesC,
                         tokenC,
                         owner,
                         flags & ~UNIQUE_TOKEN);
        return(1);
    }
    PrintMsg("attempted tpf_rsysc twice with the same address");
    rc = ReturnSystemHeap(systemHeapPtrA, nbrOfFramesA, tokenA, owner,
                          (flags | ERROR_PATH) & ~UNIQUE_TOKEN);
    if (rc == RSYSC_OK) {
        PrintError("tpf_rsysc failed to fail on double release");
        ReturnSystemHeap(systemHeapPtrB,
                         nbrOfFramesB,
                         tokenB,
                         owner,
                         flags);
        ReturnSystemHeap(systemHeapPtrC,
                         nbrOfFramesC,
                         tokenC,
                         owner,
                         flags & ~UNIQUE_TOKEN);
        return(1);
    }

    // done with systemHeapPtrB
    rc = ReturnSystemHeap(systemHeapPtrB,
                          nbrOfFramesB,
                          tokenB,
                          owner,
                          flags);

    if ((USE_UNIQUE_TOKEN) && (rc == RSYSC_OK)) {
    // error test case: release same unique token twice
        PrintMsg("attempted tpf_rsysc twice with the same unique token");
        rc = ReturnSystemHeap(systemHeapPtrB,
                              nbrOfFramesB,
                              tokenB,
                              owner,
                              flags|ERROR_PATH);
        if (rc == RSYSC_OK) {
            PrintError("tpf_rsysc failed to fail double return of unique token");
            ReturnSystemHeap(systemHeapPtrC,
                             nbrOfFramesC,
                             tokenC,
                             owner,
                             flags & ~UNIQUE_TOKEN);
            return(1);
        }
    // error test case: find returned token
    //                  (before matching non-unique token is returned)
        PrintMsg("attempted tpf_fsysc with returned token");
        if (FindSystemHeap(tokenB, systemHeapPtrB, flags|ERROR_PATH) == 0) {
            sprintf(msgBuffer,
                "tpf_fsysc failed to fail with returned token \"%s\"",
                tokenB);
            PrintError(msgBuffer);
            ReturnSystemHeap(systemHeapPtrC,
                             nbrOfFramesC,
                             tokenC,
                             owner,
                             flags & ~UNIQUE_TOKEN);
            return(1);
        }
    }

    // done with systemHeapPtrC
    ReturnSystemHeap(systemHeapPtrC,
                     nbrOfFramesC,
                     tokenC,
                     owner,
                     flags & ~UNIQUE_TOKEN);

    return(0);
}

void *GetSystemHeap(int nbrOfFrames, char *tokenID, char *ownerID, int flags)
{
    if (USE_TPF51_API) {
        return(GetSystemHeapTPF51API(nbrOfFrames, tokenID, ownerID, flags));
    } else {
        return(GetSystemHeapTPF41API(nbrOfFrames, tokenID, flags));
    }
}

void *GetSystemHeapTPF51API(int nbrOfFrames,
                            char *tokenID,
                            char *ownerID,
                            int flags)
{
    void *systemHeapPtr;
    char *ownerString;
    int   gsysc_flags = 0;
    char  flagString[50]  = "";
    char  tokenString[57] = "NULL";

    errno = 0; // clear any previous errors

    // set up indicators for gsysc and subsequent message printing
    if (USE_1MEG_FRAMES) {
        gsysc_flags |= GSYSC_1MB;
        strcpy(flagString, "GSYSC_1MB");
    }

    if (USE_64BIT) {
        gsysc_flags |= GSYSC_64BIT;
        if (flagString[0]) {
            strcat(flagString, "|64BIT");
        } else {
            strcpy(flagString, "GSYSC_64BIT");
        }
    }

    if (USE_UNIQUE_TOKEN) {
        gsysc_flags |= GSYSC_UNIQUE;
        if (flagString[0]) {
            strcat(flagString, "|UNIQUE");
        } else {
            strcpy(flagString, "GSYSC_UNIQUE");
        }
    }

    if (flagString[0] == '\0') {
        strcpy(flagString, "no flags");
    }

    if (tokenID) {
        BuildTokenString(tokenID, tokenString);
    }

    if (ownerID) {
        ownerString = "owner";
    } else {
        ownerString = "NULL";
    }

    // do the actual tpf_gsysc call
    systemHeapPtr = tpf_gsysc(nbrOfFrames, tokenID, ownerID, gsysc_flags);

    // display an informational message
    if (systemHeapPtr) {
        sprintf(msgBuffer,
            "Successful tpf_gsysc(%i, %s, %s, %s) address is %lX",
            nbrOfFrames, tokenString, ownerString, flagString, (long)systemHeapPtr);
        PrintMsg(msgBuffer);
    } else {
        if (!EXPECTED_ERROR) {
            sprintf(msgBuffer,
                "Unsuccessful tpf_gsysc(%i, %s, %s, %s)",
                nbrOfFrames, tokenString, ownerString, flagString);
            PrintError(msgBuffer);
        }
    }

    if (!EXPECTED_ERROR) {
        prterrno(errno);
    }

    return(systemHeapPtr);
}

void *GetSystemHeapTPF41API(int nbrOfFrames, char *tokenID, int flags)
{
    void *systemHeapPtr;
    char  tokenString[57] = "NULL";

    errno = 0; // clear any previous errors

    // do the actual gsysc call
    systemHeapPtr = gsysc(nbrOfFrames, tokenID);

    // display an informational message
    if (tokenID) {
        BuildTokenString(tokenID, tokenString);
    }
    if (systemHeapPtr) {
        sprintf(msgBuffer,
                "Successful gsysc(%i, %s) address is %lX",
                nbrOfFrames, tokenString, (long)systemHeapPtr);
        PrintMsg(msgBuffer);
    } else {
        if (!EXPECTED_ERROR) {
            sprintf(msgBuffer,
                    "Unsuccessful gsysc(%i, %s)", nbrOfFrames, tokenString);
            PrintError(msgBuffer);
        }
    }

    if (!EXPECTED_ERROR) {
        prterrno(errno);
    }

    return(systemHeapPtr);
}

int FindSystemHeap(char *tokenID, void *systemHeapPtr, int flags)
{
    void *foundHeapPtr;
    char  tokenString[57] = "NULL";

    if (!USE_UNIQUE_TOKEN) {
       // nothing to do if non-unique token
       if (EXPECTED_ERROR) {
          return(1);
       } else {
          return(0);
       }

    }


    if (tokenID > 0) {
        BuildTokenString(tokenID, tokenString);
       }

    errno = 0; // clear any previous errors


    foundHeapPtr = tpf_fsysc(tokenID, NULL);



    // if it worked, print a message and return zero
   if ((foundHeapPtr == systemHeapPtr) && (errno == 0)){
        sprintf(msgBuffer,
                "Successful tpf_fsysc(%s) address is %lX",
                tokenString, (long)foundHeapPtr);
        PrintMsg(msgBuffer);

        return(0);
    }


    // if it wasn't expected to work, just quietly return non-zero
    if (EXPECTED_ERROR) {
        return(1);
    }

    if (foundHeapPtr == 0) {
        // if it didn't find any address say it failed
        sprintf(msgBuffer,
                "tpf_fsysc(%s) returned zero",
                tokenString);
    } else {
        // otherwise say it returned the wrong address
        sprintf(msgBuffer,
                "tpf_fsysc(%s) returned %lX but %lX was expected",
                tokenString, (long)foundHeapPtr, (long)systemHeapPtr);
    }
    PrintError(msgBuffer);
    prterrno(errno);

    return(1);
}


int ReturnSystemHeap(void *systemHeapPtr,
                     int nbrOfFrames,
                     char *tokenID,
                     char *ownerID,
                     int flags)
{
    int rc;
    char *rcString;
    char  tokenString[57] = "NULL";

    errno = 0; // clear any previous errors

    if (tokenID) {
        BuildTokenString(tokenID, tokenString);
    }

    if (USE_UNIQUE_TOKEN) {
        // use TPF51 unique token API
        rc = tpf_rsysc(NULL, 0, tokenID);
        if (rc == RSYSC_OK) {
            sprintf(msgBuffer,
                    "Successful tpf_rsysc(NULL, NULL, %s)",
                    tokenString);
            PrintMsg(msgBuffer);
        } else {
            if ((!EXPECTED_ERROR) || (rc != RSYSC_ERROR)) {
                if (rc == RSYSC_ERROR) {
                    rcString = "RSYSC_ERROR";
                } else {
                    rcString = "not recognized";
                }
                sprintf(msgBuffer,
                        "Unsuccessful tpf_rsysc(NULL, NULL, %s) "
                           "return code is %s",
                        tokenString, rcString);
                PrintError(msgBuffer);
            }
        }
    } else {
        if (USE_TPF51_API) {
            // use TPF51 non-unique token API
            rc = tpf_rsysc(systemHeapPtr, nbrOfFrames, tokenID);
            if (rc == RSYSC_OK) {
                sprintf(msgBuffer,
                        "Successful tpf_rsysc(%lX, %i, %s)",
                        (long)systemHeapPtr, nbrOfFrames, tokenString);
                PrintMsg(msgBuffer);
            } else {
                if ((!EXPECTED_ERROR) || (rc != RSYSC_ERROR)) {
                    if (rc == RSYSC_ERROR) {
                        rcString = "RSYSC_ERROR";
                    } else {
                        rcString = "not recognized";
                    }
                    sprintf(msgBuffer,
                        "Unsuccessful tpf_rsysc(%lX, %i, %s) return code is %s",
                        (long)systemHeapPtr, nbrOfFrames, tokenString, rcString);
                    PrintError(msgBuffer);
                }
            }
       } else {
            // use TPF41 API
            rc = rsysc(systemHeapPtr, nbrOfFrames, tokenID);
            if (rc == RSYSC_OK) {
                sprintf(msgBuffer,
                        "Successful rsysc(%lX, %i, %s)",
                        (long)systemHeapPtr, nbrOfFrames, tokenString);
                PrintMsg(msgBuffer);
            } else {
                if ((!EXPECTED_ERROR) || (rc != RSYSC_ERROR)) {
                    if (rc == RSYSC_ERROR) {
                        rcString = "RSYSC_ERROR";
                    } else {
                        rcString = "not recognized";
                    }
                    sprintf(msgBuffer,
                        "Unsuccessful rsysc(%lX, %i, %s) return code is %s",
                        (long)systemHeapPtr, nbrOfFrames, tokenString, rcString);
                    PrintError(msgBuffer);
                }
            }
       }
    }

    if (!EXPECTED_ERROR) {
        prterrno(errno);
    }

    return(rc);
}

int  WriteToSystemHeap(void *systemHeapPtr, int nbrOfFrames, int flags)
{
    char *lastByte;
    size_t  size;

    if (USE_1MEG_FRAMES) {
        size = nbrOfFrames * ONE_MEG;
    } else {
        size = nbrOfFrames * FOUR_KB;
    }
    lastByte = (char *)systemHeapPtr + size - 1;
    cinfc(CINFC_WRITE,CINFC_CMMCTK2);
    *lastByte = 0x55;
    if (*lastByte != 0x55) {
       keyrc();
       PrintError("Unable to set last byte of system heap to 0x55");
       return(1);
    } else {
       memset(systemHeapPtr, 0x00, size);
       keyrc();
       if (*(lastByte) != 0x00) {
           PrintError("Unable to memset system heap to nulls");
           return(1);
       }
    }

    sprintf(msgBuffer, "Successfully wrote to %li bytes of system heap", size);
    PrintMsg(msgBuffer);

    return(0);
}

void prterrno(int nbr)
{
    char *def, *desc;

    // just return if there's no error
    if (nbr == 0) {
        return;
    }

    desc = strerror(nbr);

    switch(nbr) {
    case ENOMEM:
        def = " (ENOMEM)";
        break;
    case EINVAL:
        def = " (EINVAL)";
        break;
    default:
        def = "";
    }

    if (*desc) {
        sprintf(msgBuffer, "errno is %i%s: %s", nbr, def, desc);
    } else {
        sprintf(msgBuffer, "errno is %i%s", nbr, def);
    }
    PrintError(msgBuffer);
}

void PrintMsg(char *msg)
{
    char myBuffer[230];

    sprintf(myBuffer, "%s: %s\n", msgPrefix, msg);
    prtmsg(myBuffer);

}

void PrintError(char *msg)
{
    char myBuffer[230];

    sprintf(myBuffer, "%s: %s %s\n", msgPrefix, "***Case Failed***", msg);
    prterr(myBuffer);

}

void BuildTokenString (char *src, char *dest)
{
    char *nullSymbol = "(null)";
    char *allNulls   = "\"(all nulls)\"";
    int  i, j, isAllNulls, isMinus1;
    char *minus1 = "-1";

    isAllNulls = TRUE;
    isMinus1   = FALSE;;
    j = 0;

    // initial quote
    dest[j++] = '"';

    // move in token characters, noting nulls
    for (i = 0; i < 8; i++) {
        // if the token src addr is -1 (for one of the error test cases),
        // set up the flag and break out from this for loop.
        if (src == (void *) -1)
        {
           isMinus1 = TRUE;
           break;
        }
        else if (src[i]) {
            dest[j++] = src[i];
            isAllNulls = FALSE;
        } else {
            memcpy(&dest[j], nullSymbol, strlen(nullSymbol));
            j += strlen(nullSymbol);
        }
    }

    // note null terminator, if present
    if (!isMinus1 && !src[i]) {
        memcpy(&dest[j], nullSymbol, strlen(nullSymbol));
        j += strlen(nullSymbol);
    }

    // trailing quote
    dest[j++] = '"';
    dest[j] = '\0';

    // set the token string as "-1"
    if (isMinus1 == TRUE) {
        memcpy(dest, minus1, strlen(minus1)+1);
    }

    // chuck it all and just say "all nulls" if token is all nulls
    if (isAllNulls == TRUE) {
        memcpy(dest, allNulls, strlen(allNulls)+1);
    }
 }
