/*********************************************************************/
/*      Licensed Materials - Property of IBM                         */
/*      "Restricted Materials of IBM"                                */
/*      5748-T15                                                     */
/*      (C) Copyright IBM Corp. 2006                                 */
/*      All Rights Reserved                                          */
/*      US Government Users Restricted Rights - Use, duplication     */
/*      or dislosure restricted by GSA ADP Schedule Contract         */
/*      with IBM Corp.                                               */
/*********************************************************************/
/* This file is provided on an "as is" basis.  There are no          */
/* warranties, express or implied, including the implied warranties  */
/* of merchantability and fitness for a particular purpose.          */
/* IBM has no obligation to provide service, defect correction, or   */
/* any maintenance for the file.  IBM has no obligation to supply    */
/* any updates or enhancements for the file to you even if such are  */
/* or later become available.                                        */
/*********************************************************************/
/* Change History                                                    */
/*                                                                   */
/*                                                                   */
/*                                                                   */
/*********************************************************************/
/* GENERAL INFORMATION:                                              */
/*                                                                   */
/* Segment Name:  qm5b.c                                             */
/*                                                                   */
/* Abstract:      Member of the QM5B CSO which is used by the QM5A   */
/*                TPF GCC MEM5 driver DLM.                           */
/*                                                                   */
/* Description:   This segment contains MEM5 driver tests for        */
/*                ECB Heap & System Heap that require a second ECB.  */
/*                                                                   */
/*********************************************************************/
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <tpf/tpfapi.h>
#include <tpf/c_eb0eb.h>
#include <tpf/sysapi.h>
#include <errno.h>
#include "qm5a70.h"

// prototypes:
void UseSystemHeap(struct SystemHeapParms *);
void FreeECBHeap(int);
void ReallocECBHeap(int);
void MallocECBHeap(int);
void MAXC_func(int);
void Get_getcc(int size);

char msgBuffer[230];
char *msg_prefix = "QM5B";

void QM5B(void)
{
    struct PassedParms *p_ptr, parms;
    char *input, *output;
    int i, j;

    // Change the input "string" back into hex data
    input = ecbptr()->ce1cr0;
    output = (char *)&parms;
    for (i = 0, j = 0; i < sizeof(struct PassedParms); i++, j += 2) {
        output[i] = input[j] & input[j+1];
    }
    p_ptr = &parms;
    errorFlag =   p_ptr->errorFlagSetting;
    verboseFlag = p_ptr->verboseFlagSetting;


    switch (p_ptr->task) {
        case USE_SYSTEM_HEAP:
            UseSystemHeap((struct SystemHeapParms *)p_ptr->genericPtr);
            break;
        case FREE_ECB_HEAP:
            FreeECBHeap (p_ptr->genericNum);
            break;
        case REALLOC_ECB_HEAP:
            ReallocECBHeap (p_ptr->genericNum);
            break;
       case MALLOC_ECB_HEAP:
            MallocECBHeap (p_ptr->genericNum);
            break;
       case MAX_ECB:
            MAXC_func(p_ptr->genericNum);
            break;
        default:
            PrintError("QM5B: unknown task");
            break;
    }

    PrintError("QM5B: failed to fail");
}


void FreeECBHeap(int testNumber)
{
    void *storage1_ptr = NULL;
    void *storage2_ptr = NULL;


    switch (testNumber) {

        case 1:        // test case : realloc, free original ptr
            PrintMsg ("attempting free of original realloc ptr");
            storage1_ptr = (char *)malloc(1000);
            if ((storage2_ptr = (char *)realloc(storage1_ptr, 5000))) {
                free (storage1_ptr);            // DUMPS
            }
            sprintf (msgBuffer, "free did not fail with original realloc ptr %lX", (long)storage1_ptr);
            PrintMsg(msgBuffer);
            break;

        case 2:            // test case : free invalid address
            PrintMsg("attempting free with address of -1");
            free ((void*)-1);
            sprintf (msgBuffer, "free did not fail with invalid address %lX", (long)-1);
            PrintMsg(msgBuffer);
            break;

        case 3:            // test case : free embedded address
            PrintMsg("attempting free of address inside malloc address");
            storage1_ptr = (char *)malloc(10);
            storage1_ptr = (char *)storage1_ptr + 1;
            free (storage1_ptr);
            sprintf (msgBuffer, "free did not fail with embedded address %lX", (long)storage1_ptr);
            PrintMsg(msgBuffer);
            break;

        default:
            PrintError("QM5B: unknown test number");
            break;
    }

    exit(0);

}

void ReallocECBHeap(int testNumber)
{
    char *realloc_ptr = NULL;
    void *malloc_ptr = NULL;
    char *invalid_ptr = (char *)-1;
    int  size = 1000;

    switch (testNumber) {

        case 1:        // test case : realloc with invalid address
            //   use 4.1 realloc api
            PrintMsg ("attempting realloc with invalid address");
            if ((realloc_ptr = (char *)realloc (invalid_ptr, size))) {
                sprintf (msgBuffer, "realloc did not fail with invalid ptr %lX", (long)invalid_ptr);
                PrintMsg(msgBuffer);
                exit(0);
            } else {
                  sprintf (msgBuffer, "Unsuccessful realloc: addr %lX %i bytes", (long)realloc_ptr, size);
                  PrintError(msgBuffer);
                  prterrno(errno);
            }

            //   use zTPF realloc api
            PrintMsg ("attempting realloc64 with invalid address");
            if ((realloc_ptr = (char *) realloc64 (invalid_ptr, size))) {
                sprintf (msgBuffer, "realloc64 did not fail with invalid ptr %lX", (long)invalid_ptr);
                PrintMsg(msgBuffer);
                exit(0);
            } else {
                  sprintf (msgBuffer, "Unsuccessful realloc64: addr %lX %i bytes", (long)realloc_ptr, size);
                  PrintError(msgBuffer);
                  prterrno(errno);
            }

        case 2:        // test case : malloc, free, realloc with freed ptr
            //   use 4.1 realloc api
            PrintMsg ("attempting realloc with freed ptr");
            malloc_ptr = malloc(1000);
            free(malloc_ptr);
            if ((realloc_ptr = (char *)realloc (malloc_ptr, size))) {
                sprintf (msgBuffer, "realloc did not fail with freed ptr %lX", (long)malloc_ptr);
                PrintMsg(msgBuffer);
                exit(0);
            } else {
                  sprintf (msgBuffer, "Unsuccessful realloc: addr %lX %i bytes", (long)realloc_ptr, size);
                  PrintError(msgBuffer);
                  prterrno(errno);
                  free(realloc_ptr);
            }

            //   use zTPF realloc api
            PrintMsg ("attempting realloc64 with freed ptr");
            malloc_ptr = malloc(1000);
            free(malloc_ptr);
            if ((realloc_ptr = (char *) realloc64(malloc_ptr, size))) {
                sprintf (msgBuffer, "realloc64 did not fail with freed ptr %lX", (long)malloc_ptr);
                PrintMsg(msgBuffer);
                exit(0);
            } else {
                  sprintf (msgBuffer, "Unsuccessful realloc64: addr %lX %i bytes", (long)realloc_ptr, size);
                  PrintError(msgBuffer);
                  prterrno(errno);
                  free(realloc_ptr);
            }

        default:
            PrintError("QM5B: unknown test number");
            break;
    }

    PrintMsg("ReallocECBHeap exit");
    exit(0);
}

void MallocECBHeap(int testNumber)
{

    void *malloc_ptr = NULL;
    char *last_byte;
    int size, value = 0xAB;

    switch (testNumber) {
                       // 64b-48: Test integrity check, should get CTL-75
        case 1:        // test case using free(), for 8-byte alignment- need 272
            size = 270;
            malloc_ptr = malloc(size);
            last_byte = (char *)malloc_ptr + size + 2;
            PrintMsg("attempting to write past the end of malloc area, should get CTL-75");
            *(last_byte) = value;
            free(malloc_ptr);
            PrintError("Did not get CTL-75 ");
            break;

        case 2:        // test case using realloc( )
            size = 650;
            malloc_ptr = malloc(size);
            PrintMsg("attempting to write past the end of malloc area, should get CTL-75");
            *((char *)malloc_ptr + size + 13) = 0xAC;  // 6 to alignment, 7 to last byte
            realloc(malloc_ptr, 0);
            PrintError("Did not get CTL-75 ");
            break;

        default:
            PrintError("QM5B: unknown test number");
            break;
    }
    exit(0);
}



void UseSystemHeap(struct SystemHeapParms *systemHeapPtr)
{
    void *foundHeapPtr;
    int   nbrOfFrames, flags, rc;
    char  tokenID[9];

    // read from system heap from another ECB
    cinfc(CINFC_WRITE,CINFC_CMMCTK2);
    nbrOfFrames = systemHeapPtr->nbrOfFrames;
    flags = systemHeapPtr->flags;
    memcpy(tokenID, systemHeapPtr->token, sizeof(tokenID)-1);
    tokenID[8] = '\0'; // null terminate token
    sprintf(msgBuffer,
            "Successfully read from system heap at %lX",
            (long)systemHeapPtr);
    PrintMsg(msgBuffer);

    // write to sytem heap from another ECB
    systemHeapPtr->nbrOfFrames = -1;
    if (systemHeapPtr->nbrOfFrames != -1) {
        keyrc();
        sprintf(msgBuffer,
                "Unable to write to system heap at %lX",
                (long)systemHeapPtr);
        PrintError(msgBuffer);
        if (DoErrors) {
            PrintError("Unable to run error tests");
        }
        exit(1);
    } else {
        keyrc();
        sprintf(msgBuffer,
                "Successfully wrote to system heap at %lX",
                (long)systemHeapPtr);
        PrintMsg(msgBuffer);
    }

    // find system heap token from another ECB
    if ((USE_UNIQUE_TOKEN) && (!DoErrors)) {
        foundHeapPtr = tpf_fsysc(tokenID, NULL);
        if (foundHeapPtr == systemHeapPtr) {
            sprintf(msgBuffer,
                    "Successful tpf_fsysc(%s), address is %lX",
                    tokenID, (long)foundHeapPtr);
            PrintMsg(msgBuffer);
        } else {
            sprintf(msgBuffer,
                   "tpf_fsysc(%s) returned %lX but %lX was expected",
                   tokenID, (long)foundHeapPtr, (long)systemHeapPtr);
           PrintError(msgBuffer);
           exit(1);
       }
   }

   // return system heap from another ECB
   rc = tpf_rsysc(systemHeapPtr, nbrOfFrames, tokenID);
   if (rc == RSYSC_OK) {
       sprintf(msgBuffer, "Successful tpf_rsysc(%lX, %i, token)",
               (long)systemHeapPtr, nbrOfFrames);
       PrintMsg(msgBuffer);
   } else {
        sprintf(msgBuffer, "Unsuccessful tpf_rsysc(%lX, %i, token)",
                (long)systemHeapPtr, nbrOfFrames);
        PrintError(msgBuffer);
        if (DoErrors) {
            PrintError("Unable to run error tests");
        }
        exit(1);
    }

    exit(0);
}

void MAXC_func(int testNumber)
{
       sleep(1);  // allow output messages to catch up so they are in the right place
       int maxPA = ecbp2()->ce2epriv;      // maximum storage for ECB Private Area
       int rc;

       /*************************************/
       /* Parmlists for tpf_ebmaxc function */
       /*************************************/

       // raise ECB maximum values by less than or equal to 0
       struct ebmaxc_parmlist myparm1[] =  {
               { MAX31HEAP ,  -4 },
               { MAXEPRIV  ,  -2 },
               { MAX64HEAP ,   0 },
               { 0, 0 }
       };

       // raise only ECB Private Area to the maximum level
       struct ebmaxc_parmlist myparm2[] =  {
               { MAXEPRIV  ,   MAXVALUE },
               { 0, 0 }
       };

       // rase only ECB Private Area by more than maximum
       struct ebmaxc_parmlist myparm3[] =  {
               { MAXEPRIV  ,  (MAXVALUE+1) },
               { 0, 0 }
       };


       PrintMsg("Entering QM5B MAXC_func");

       switch (testNumber) {
               // run normal cases for ECB Private Area
               case 1 :
                       /***********************/
                       /* use system default  */
                       /***********************/

                       // allocate space from the ECB Private Area
                       Get_getcc(maxPA);

                       /*******************************/
                       /* increase max using MAXVALUE */
                       /*******************************/

                       // raise maximum to the most the system will allow
                       tpf_ebmaxc(myparm2);

                       maxPA = ecbp2()->ce2epriv;      // maximum storage for ECB Private Area

                       // allocate space from the ECB Private Area
                       Get_getcc(maxPA);

                       /******************************************/
                       /* increase max more than the system max  */
                       /******************************************/

                       // raise ECB Private area maximum values by more than the system will allow
                       // the value should not be greater than the system maximum
                       tpf_ebmaxc(myparm3);

                       if(maxPA != (rc = ecbp2()->ce2epriv))
                       {
                               sprintf(msgBuffer, "case 1 : Private Area max changed from %d (MB) to %d (MB)", maxPA, rc);
                               PrintError(msgBuffer);
                       }

                       PrintMsg("case 1: completed successfully");
               break;

               // run error case for ECB Private Area
               case 2 :
                       PrintMsg("*** Attempting to allocate greater than ECB Private Area max - should cause a CTL - 6DB ***\n");
                       Get_getcc(maxPA+1);
                       PrintError("case 2: getcc failed to fail with greater than max values");
               break;

               // call myparm1 to raise maximum ECB values by less than or equal to 0
               // should result in a dump
               case 3 :
                       tpf_ebmaxc(myparm1);
                       PrintError("case 3: tpf_ebmaxc failed to fail with negative values");
               break;

               default:
                       PrintError("QM5B: unknown test number");
               break;
       }


       PrintMsg("QM5B: exiting MAXC_func\n");
       sleep(1);
       exit(0);
}

void Get_getcc(int size)
{
       int level;      // data level
       int rc, counter;

       char *work1;            // pointer to get blocks

       // release level if block already exists on it
       for(rc = 0; rc < size; rc++)
       {
               level = rc * 8;
               if(levtest(level))
                       relcc(level);
       }

       // get blocks
       for(rc = 0; rc < size; rc++)
       {
               level = rc * 8;
               counter = 0;
               // get the data block
               while(counter < 240)
               {
                       // get the data block
                       work1 = (char *) getcc(level, GETCC_TYPE, L4);
                       // detach the data block so you can get another on the same level
                       detac(level);
                       counter++;
               }
               sprintf(msgBuffer, "Successfully getcc and detac 240 times for data level %d\n", rc);
               PrintMsg(msgBuffer);
       }

       // release all levels
       for(rc = 0; rc < size; rc++)
       {
               counter = 0;
               level = rc * 8;

               while(counter < 240)
               {
                       attac(level);
                       relcc(level);
                       counter++;
               }
               sprintf(msgBuffer, "Successfully attac and relcc 240 times for data level %d\n", rc);
               PrintMsg(msgBuffer);
       }

       return;
}

void PrintMsg(char *msg)
{
    char msgBuffer[230];

    sprintf(msgBuffer, "%s: %s", msg_prefix, msg);
    prtmsg(msgBuffer);
}

void PrintError(char *msg)
{
    char msgBuffer[230];

    sprintf(msgBuffer, "%s: %s %s", msg_prefix, "***Case Failed***", msg);
    prterr(msgBuffer);
}

void prterrno(int nbr)
{
    char *def, *desc;

    // just return if there's no error
    if (nbr == 0) {
        return;
    }

    desc = strerror(nbr);

    switch(nbr) {
    case ENOMEM:
        def = " (ENOMEM)";
        break;
    case EINVAL:
        def = " (EINVAL)";
        break;
    default:
        def = "";
    }

    if (*desc) {
        sprintf(msgBuffer, "errno is %i%s: %s", nbr, def, desc);
    } else {
        sprintf(msgBuffer, "errno is %i%s", nbr, def);
    }
    PrintError(msgBuffer);
}
